import logging

from monopoly.config import PdfConfig, StatementConfig, settings
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
)

from ..base import ProcessorBase

logger = logging.getLogger(__name__)


class Hsbc(ProcessorBase):
    credit_config = StatementConfig(
        bank_name=BankNames.HSBC,
        statement_date_pattern=r"Statement From (\d{2}\s[A-Z]{3}\s\d{4}) to",
        prev_balance_pattern=StatementBalancePatterns.HSBC,
        statement_date_format=r"%d %b %Y",
        transaction_pattern=CreditTransactionPatterns.HSBC,
        transaction_date_format="%d %b",
        multiline_transactions=True,
    )

    debit_config = None

    pdf_config = PdfConfig(
        passwords=settings.hsbc_pdf_passwords,
        page_bbox=(0, 0, 379, 842),
    )

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.6, algorithm=4, revision=4, length=128, permissions=-1804
        ),
        MetadataIdentifier(
            title="PRJ_BEAGLE_ST_CNS_SGH_APP_Orchid",
            creator="OpenText Exstream",
        ),
    ]



from typing import Annotated, Optional

from pydantic import StringConstraints
from pydantic.dataclasses import dataclass
from pydantic_settings import BaseSettings, SettingsConfigDict

from monopoly.constants import BankNames


class Settings(BaseSettings):
    """
    Pydantic model that automatically populates variables from a .env file.
    """

    ocbc_pdf_passwords: Optional[list[str]] = None
    citibank_pdf_passwords: Optional[list[str]] = None
    standard_chartered_pdf_passwords: Optional[list[str]] = None
    hsbc_pdf_passwords: Optional[list[str]] = None

    model_config = SettingsConfigDict(env_file=".env", extra="allow")


# pylint: disable=too-many-instance-attributes
@dataclass
class StatementConfig:
    """
    Dataclass storing configuration values for debit and credit card statements

    - `transaction_pattern` refers to the regex pattern used to capture transactions,
    where a pattern like:
        "(?P<transaction_date>\\d+/\\d+)\\s*"
        "(?P<description>.*?)\\s*"
        "(?P<amount>[\\d.,]+)$"
    is used to capture a transaction like:
        06/07 URBAN TRANSIT CO. SINGAPORE SG  1.38
    - `transaction_date_format` represents the datetime format that a specific bank uses
    for transactions. For example, "%d/%m" will match 06/07
    - `multiline_transactions` controls whether Monopoly tries to concatenate
    transactions that are split across two lines
    - `debit_account_identifier` is a regex pattern that is used to determine whether
    a statement from a bank is a debit or credit card statement.
    - `prev_balance_pattern` is a regex pattern used to match the previous balance
    line in a bank statement, which is later converted into a transaction.
    """

    bank_name: BankNames
    transaction_pattern: str
    transaction_date_format: Annotated[str, StringConstraints(pattern="%")]
    statement_date_pattern: str
    statement_date_format: Annotated[str, StringConstraints(pattern="%.+%.+%")]
    multiline_transactions: bool = False
    debit_account_identifier: Optional[str] = None
    prev_balance_pattern: Optional[str] = None

    # Convert enums to strings
    def __post_init__(self):
        self.bank_name = self.bank_name.value


@dataclass
class PdfConfig:
    """Stores PDF configuration values for the `PdfParser` class

    - `password`: The password used to unlock the PDF (if it is locked)
    - `page_range`: A slice representing which pages to process. For
    example, a range of (1, -1) will mean that the first and last pages
    are skipped.
    - `page_bbox`: A tuple representing the bounding box range for every
    page. This is used to avoid weirdness like vertical text, and other
    PDF artifacts that may affect parsing.
    """

    passwords: Optional[list[str]] = None
    page_range: tuple[Optional[int], Optional[int]] = (None, None)
    page_bbox: Optional[tuple[float, float, float, float]] = None


settings = Settings()



from datetime import datetime
from hashlib import sha256
from pathlib import Path

from monopoly.config import StatementConfig


def generate_hash(file_path: Path) -> str:
    """
    Generates a hash of the raw PDF content
    """
    with open(file_path, "rb") as file:
        file_content = file.read()
        hash_object = sha256()
        hash_object.update(file_content)
        file_hash = hash_object.hexdigest()[0:6]
    return file_hash


def generate_name(
    file_path: Path,
    format_type: str,
    config: StatementConfig,
    account_type: str,
    statement_date: datetime,
) -> str:
    """
    Generates a new file name, depending on: bank, account type, statement date,
    and old file name.

    e.g. 20230720.pdf -> hsbc-credit-2023-06-b960bf1e.csv

    The appeneded UUID ensures that two statements from the same bank in a
    month will not overwrite each other.
    """
    bank_name = config.bank_name
    year = statement_date.year
    month = statement_date.month
    file_uuid = generate_hash(file_path)

    file_suffix = "csv"

    filename = (
        f"{bank_name}-{account_type}-{year}-{month:02d}-{file_uuid}.{file_suffix}"
    )

    if format_type == "blob":
        return (
            f"bank_name={bank_name}/"
            f"account_type={account_type}/"
            f"year={year}/"
            f"month={month}/"
            f"filename"
        )

    if format_type == "file":
        return filename

    raise ValueError("Invalid format_type")



from enum import StrEnum, auto

from pydantic.dataclasses import dataclass


# pylint: disable=unused-argument,no-self-argument
class AutoEnum(StrEnum):
    """Generates lower case values for enums
    e.g. CITIBANK -> citibank
    """

    def _generate_next_value_(name: str, *_):  # type: ignore
        return name.lower()


class AccountType(AutoEnum):
    CREDIT = auto()
    DEBIT = auto()


class BankNames(AutoEnum):
    CITIBANK = auto()
    EXAMPLE = auto()
    DBS = auto()
    HSBC = auto()
    OCBC = auto()
    STANDARD_CHARTERED = auto()


class StatementFields(AutoEnum):
    TRANSACTION_DATE = auto()
    DESCRIPTION = auto()
    AMOUNT = auto()


class SharedPatterns(StrEnum):
    """
    AMOUNT matches the following patterns:
    1,123.12 | 123.12 | (123.12) | ( 123.12) | 123.12 CR
    """

    AMOUNT = r"(?P<amount>[\d.,]+)\s*"
    AMOUNT_EXTENDED = r"(?P<amount>[\d.,]+|\([\d.,\s]+\)$)\s*(?P<suffix>CR|DR)?$"
    AMOUNT_EXTENDED_WITHOUT_EOL = (
        r"(?P<amount>[\d.,]+|\([\d.,\s]+\)$)\s*(?P<suffix>CR|DR)?"
    )
    AMOUNT_WITH_CASHBACK = r"(?P<amount>[\d.,]+|\([\d.,\s]+\))$"
    BALANCE = r"(?:(?P<balance>[\d.,]+)\s*)?$"
    DESCRIPTION = r"(?P<description>.*?)\s+"
    TRANSACTION_DATE_ABBREVIATED_ALL_CAPS = r"(?P<transaction_date>\d{2}\s[A-Z]{3})\s+"
    TRANSACTION_DATE_ABBREVIATED_PROPER_CASE = (
        r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
    )
    POSTING_DATE_ABBREVIATED_PROPER = r"(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"


class StatementBalancePatterns(StrEnum):
    DBS = (
        r"(?P<description>PREVIOUS BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    CITIBANK = (
        r"(?P<description>BALANCE PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    HSBC = (
        r"(?P<description>Previous Statement Balance?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    OCBC = (
        r"(?P<description>LAST MONTH'S BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    STANDARD_CHARTERED = (
        r"(?P<description>BALANCE FROM PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )


class CreditTransactionPatterns(StrEnum):
    DBS = (
        SharedPatterns.TRANSACTION_DATE_ABBREVIATED_ALL_CAPS
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    CITIBANK = (
        SharedPatterns.TRANSACTION_DATE_ABBREVIATED_ALL_CAPS
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_WITH_CASHBACK
    )
    HSBC = (
        SharedPatterns.POSTING_DATE_ABBREVIATED_PROPER
        + SharedPatterns.TRANSACTION_DATE_ABBREVIATED_PROPER_CASE
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    OCBC = (
        r"(?P<transaction_date>\d+/\d+)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_WITH_CASHBACK
    )
    STANDARD_CHARTERED = (
        SharedPatterns.POSTING_DATE_ABBREVIATED_PROPER
        + SharedPatterns.TRANSACTION_DATE_ABBREVIATED_PROPER_CASE
        + SharedPatterns.DESCRIPTION
        + r"(?:(?P<transaction_ref>Transaction\sRef\s\d+)?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED
    )


class DebitTransactionPatterns(StrEnum):
    DBS = (
        SharedPatterns.TRANSACTION_DATE_ABBREVIATED_PROPER_CASE
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
        + SharedPatterns.BALANCE
    )
    OCBC = (
        SharedPatterns.TRANSACTION_DATE_ABBREVIATED_ALL_CAPS
        + r"(?P<value_date>\d{2}\s[A-Z]{3})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
        + SharedPatterns.BALANCE
    )


@dataclass
class EncryptionIdentifier:
    pdf_version: float
    algorithm: int
    revision: int
    length: int
    permissions: int


@dataclass
class MetadataIdentifier:
    title: str = ""
    author: str = ""
    subject: str = ""
    creator: str = ""
    producer: str = ""



import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

from pandas import DataFrame

from monopoly.config import PdfConfig, StatementConfig
from monopoly.constants import AccountType, StatementFields
from monopoly.pdf import PdfParser
from monopoly.statement import Statement
from monopoly.write import generate_name

logger = logging.getLogger(__name__)


class StatementProcessor(PdfParser):
    credit_config: StatementConfig
    debit_config: Optional[StatementConfig]
    pdf_config: Optional[PdfConfig] = None
    parser: Optional[PdfParser] = None
    safety_check_enabled: bool = True
    """
    Handles extract, transform and load (ETL) logic for bank statements

    Transactions are extracted from pages, before undergoing various
    transformations like converting various date formats into ISO 8601.
    Transformed statements are then written to a CSV file.

    Since the auto-detect function already creates a parser, this class
    allows for the parser to be reused and avoid re-opening the PDF.
    """

    def __init__(self, file_path: Path, parser: Optional[PdfParser] = None, **kwargs):
        keys = [
            "credit_config",
            "debit_config",
            "pdf_config",
            "safety_check_enabled",
        ]

        for key, value in kwargs.items():
            if key in keys:
                setattr(self, key, value)

        if not parser:
            super().__init__(file_path, self.pdf_config)

    def extract(self) -> Statement:
        """Extracts transactions from the statement, and performs
        a safety check to make sure that total transactions add up"""
        pages = self.get_pages()
        statement = Statement(pages, self.credit_config, self.debit_config)

        if not statement.transactions:
            raise ValueError("No transactions found - statement extraction failed")

        if not statement.statement_date:
            raise ValueError("No statement date found")

        statement = self._inject_prev_month_balance(statement)

        if self.safety_check_enabled:
            self._perform_safety_check(statement)

        return statement

    def _inject_prev_month_balance(self, statement: Statement):
        """
        Injects the previous month's balance as a transaction, if it exists
        """
        if statement.prev_month_balance:
            first_transaction_date = statement.transactions[0].transaction_date
            statement.prev_month_balance.transaction_date = first_transaction_date
            statement.transactions.insert(0, statement.prev_month_balance)
        return statement

    def _perform_safety_check(self, statement: Statement) -> bool:
        """Checks that the total sum of all transactions is present
        somewhere within the document

        Text is re-extracted from the page, as some bank-specific bounding-box
        configurations (e.g. HSBC) may preclude the total from being extracted

        Returns `False` if the total does not exist in the document.
        """
        numbers = set()
        df = statement.df
        amount = StatementFields.AMOUNT
        warning_message = (
            "Safety check failed - transactions may be missing or inaccurate"
        )
        for page in self.document:
            lines = page.get_textpage().extractText().split("\n")
            numbers.update(statement.get_decimal_numbers(lines))

        total_amount = abs(round(df[amount].sum(), 2))

        if statement.account_type == AccountType.DEBIT:
            logger.debug(
                "Total amount %s cannot be found in document - %s",
                total_amount,
                "checking to see if debit/credit sum is present in document",
            )
            debit_sum = round(abs(df[df[amount] > 0][amount].sum()), 2)
            credit_sum = round(abs(df[df[amount] < 0][amount].sum()), 2)
            result = (debit_sum in numbers) == (credit_sum in numbers)
            if not result:
                logger.warning(msg=warning_message)

        # handling for credit statement
        else:
            result = total_amount in numbers
            if not result:
                logger.warning(
                    msg=(
                        warning_message,
                        "Total amount %s cannot be found in document",
                    ),
                    args=total_amount,
                )
        return result

    def transform(self, statement: Statement) -> DataFrame:
        logger.debug("Running transformation functions on DataFrame")

        df = statement.df
        statement_date = statement.statement_date
        date_format = statement.statement_config.transaction_date_format

        def convert_date(row):
            parsed_date = datetime.strptime(
                row[StatementFields.TRANSACTION_DATE], date_format
            )
            row_year = statement_date.year
            row_day, row_month = parsed_date.day, parsed_date.month

            if statement_date.month == 1 and row_month == 12:
                row_year = statement_date.year - 1

            return f"{row_year}-{row_month:02d}-{row_day:02d}"

        logger.debug("Transforming dates to ISO 8601")
        df[StatementFields.TRANSACTION_DATE] = df.apply(convert_date, axis=1)
        return df

    def load(self, df: DataFrame, statement: Statement, output_directory: Path):
        if isinstance(output_directory, str):
            output_directory = Path(output_directory)

        filename = generate_name(
            account_type=statement.account_type,
            file_path=self.file_path,
            format_type="file",
            config=statement.statement_config,
            statement_date=statement.statement_date,
        )

        output_path = output_directory / filename
        logger.debug("Writing CSV to file path: %s", output_path)
        df.to_csv(output_path, index=False)

        return output_path



import logging
from dataclasses import Field, fields
from itertools import product
from pathlib import Path
from typing import Type

from monopoly.constants import EncryptionIdentifier, MetadataIdentifier
from monopoly.pdf import PdfParser

from .base import ProcessorBase
from .citibank import Citibank
from .dbs import Dbs
from .example_bank import ExampleBankProcessor
from .hsbc import Hsbc
from .ocbc import Ocbc
from .standard_chartered import StandardChartered

processors: list[Type[ProcessorBase]] = [
    Citibank,
    Dbs,
    ExampleBankProcessor,
    Hsbc,
    Ocbc,
    StandardChartered,
]


logger = logging.getLogger(__name__)


def detect_processor(file_path: Path) -> ProcessorBase:
    """
    Reads the encryption metadata or actual metadata (if the PDF is not encrypted),
    and checks for a bank based on unique identifiers.
    """
    parser = PdfParser(file_path)
    for processor in processors:
        metadata_items = processor.get_identifiers(parser)
        if is_bank_identified(metadata_items, processor):
            return processor(file_path=parser.file_path, parser=parser)

    raise ValueError(f"Could not find a bank for {parser.file_path}")


def is_bank_identified(
    metadata_items: list[EncryptionIdentifier | MetadataIdentifier],
    processor: Type[ProcessorBase],
) -> bool:
    """
    Checks if a bank is identified based on a list of metadata items.
    """
    for identifier, metadata in product(processor.identifiers, metadata_items):
        logger.debug(
            "Comparing bank %s identifier %s against PDF metadata %s",
            processor.__name__,
            identifier,
            metadata,
        )
        if all(
            check_matching_field(field, metadata, identifier)
            for field in fields(metadata)
        ):
            return True
    return False


def check_matching_field(
    field: Field,
    metadata: EncryptionIdentifier | MetadataIdentifier,
    identifier: EncryptionIdentifier | MetadataIdentifier,
) -> bool:
    """
    Checks if a field in the metadata matches the corresponding identifier field.
    """
    # Only compare matching identifier types
    if type(metadata) is type(identifier):
        field_value = getattr(metadata, field.name)
        identifier_value = getattr(identifier, field.name)

        # allow for partial string matching
        if isinstance(field.type(), str):
            return identifier_value in field_value

        # otherwise check that values match exactly
        return identifier_value == field_value
    return False



from pathlib import Path
from typing import Any, Optional

from monopoly.constants import EncryptionIdentifier, MetadataIdentifier
from monopoly.pdf import PdfConfig, PdfParser
from monopoly.processor import StatementProcessor


class ProcessorBase(StatementProcessor):
    """Helper class to handle initialization of common variables
    that are shared between bank processor classes"""

    identifiers: list[EncryptionIdentifier | MetadataIdentifier]

    def __init__(
        self,
        file_path: Path,
        passwords: Optional[list[str]] = None,
        parser: Optional[PdfParser] = None,
    ):
        self.parser = parser

        # optional config
        self.pdf_config = getattr(self, "pdf_config", PdfConfig())
        self.safety_check_enabled = getattr(self, "safety_check_enabled", True)

        # this allows the user to override the default pydantic password
        # and supply their own password via the bank subclasses
        if passwords:
            self.pdf_config.passwords = passwords

        if isinstance(file_path, str):
            file_path = Path(file_path)

        super().__init__(
            credit_config=self.credit_config,
            debit_config=self.debit_config,
            safety_check_enabled=self.safety_check_enabled,
            pdf_config=self.pdf_config,
            file_path=file_path,
        )

    @staticmethod
    def get_identifiers(parser: PdfParser) -> list[Any]:
        identifiers = []
        # pylint: disable=protected-access
        if parser.extractor.encrypt_dict:
            extractor = parser.extractor
            major, minor = extractor.pdf._header_version
            pdf_version = f"{major}.{minor}"
            encryption_identifier = EncryptionIdentifier(
                float(pdf_version),
                extractor.algorithm,
                extractor.revision,
                extractor.length,
                extractor.permissions,
            )
            identifiers.append(encryption_identifier)

        if metadata := parser.document.metadata:
            metadata_identifier = MetadataIdentifier(**metadata)
            identifiers.append(metadata_identifier)  # type: ignore

        if not identifiers:
            raise ValueError(f"Could not get identifier for {parser.file_path}")

        return identifiers
